# coding=utf-8
"""
@author: ltmit
"""
import sys as _sys
import time
import threading as _th
import traceback as _tb
if _sys.version[0] == '2':
    import Queue as _que
else:
    import queue as _que

class AsyncHandler(_th.Thread):
    def __init__(self, que_size=0):
        _th.Thread.__init__(self)
        self._que = _que.Queue(que_size)
        self.quit_flag = 0

    def run(self):
        try:
            while not self.quit_flag:
                obj = self._que.get()
                if obj is None: break
                self.handle(obj)
            self.texit(None)
        except Exception as e:
            self.texit(e)

    def push(self, obj, block=True):
        if block:
            self._que.put(obj)
            return True
        else:
            try:
                self._que.put_nowait(obj)
                return True
            except:
                return False

    # callbacks
    def handle(self, obj):
        pass

    def texit(self, e):
        if e:
            print(e)


class ReadWriteLock(object):
    def __init__(self):
        self.__monitor = _th.Lock()
        self.__exclude = _th.Lock()
        self.__readers = 0

    def acquire_read(self):
        with self.__monitor:
            self.__readers += 1
            if self.__readers == 1:
                self.__exclude.acquire()

    def release_read(self):
        with self.__monitor:
            self.__readers -= 1
            if self.__readers == 0:
                self.__exclude.release()

    def acquire_write(self):
        self.__exclude.acquire()

    def release_write(self):
        self.__exclude.release()

    def rlock(self):  # for 'with' statement
        class _rlw:
            def __init__(self, rw):
                self.__rw = rw

            def __enter__(self):
                self.__rw.acquire_read()

            def __exit__(self, typ, val, trac):
                self.__rw.release_read()
                return False

        return _rlw(self)

    def wlock(self):
        class _wlw:
            def __init__(self, rw):
                self.__rw = rw

            def __enter__(self):
                self.__rw.acquire_write()

            def __exit__(self, typ, val, trac):
                self.__rw.release_write()
                return False

        return _wlw(self)


'''---------------------------------------------'''


class TaskPool:
    def __init__(self, thread_ct=4, que_size=4):
        self.__que = _que.Queue(que_size)
        self.__threads = []

        def tproc(q):
            while True:
                e = q.get()
                try:
                    e.proc()
                finally:
                    q.task_done()

        while thread_ct > 0:
            t = _th.Thread(target=tproc, args=(self.__que,))
            self.__threads.append(t)
            t.setDaemon(True)
            t.start()
            thread_ct -= 1
            # print t

    def push_task(self, task, block=True):
        """task class need to implement 'proc()' method"""
        self.__que.put(task, block)

    def push_task_fb(self, func, *args):
        class _dumtask:
            def proc(self):
                try:
                    func(*args)
                except:
                    _tb.print_exc()
                    exit(1)
        self.__que.put(_dumtask(), True)

    def join(self):
        # print self.__que.qsize()
        self.__que.join()


class TimerClz(_th.Thread):
    def __init__(self, intval):
        _th.Thread.__init__(self)
        self._intval = intval

    def timer_proc(self):
        pass

    def run(self):
        while 1:
            time.sleep(self._intval)
            if self.timer_proc() == StopIteration:
                break

    def start_timer(self):
        self.setDaemon(True)
        _th.Thread.start(self)


if __name__ == '__main__':
    class testt(TimerClz):
        def __init__(self):
            TimerClz.__init__(self, 0.8)

        def timer_proc(self):
            print(time.time())


    t = testt()
    t.start_timer()
    time.sleep(11)



# coding=utf-8
import vsepy as vse
import glob
import pytensordb

if __name__=="__main__":
    print("配置连接地址")
    vc = vse.VseClient(('10.186.16.132', 2021))
    # 通过vc构造FeatureTransformer实例 ft
    ft = vse.FeatureTransformer(client=vc)
    # 打印ft参数
    print(ft)

    all_db_lst=vc.enum_all_dbs()

    '''步骤1：建库'''
    dbname = 'test_db1'
    if not dbname in all_db_lst:
        # 数据库创建
        print("创建数据库")
        vc.create_db(dbname)
        print("创建成功")
    else:
        print("库已经存在")

    # 枚举所有库
    all_db_lst=vc.enum_all_dbs()
    for dbname in all_db_lst:
        # 打印库名, 以及数据库记录数获取
        print(dbname, vc.get_db_record_count(dbname))
    
    # 从指定文件夹读取特征数据文件（二进制原始特征数据）
    rec_idx = 1
    for fp in glob.iglob('./yizhifu/*.dat'):
        with open(fp,'rb') as f:
            raw_feat = f.read()
        # 使用ft进行特征规范化转换
        feat = ft.tranform(raw_feat)
        '''
        将feat以及关联字段数据（binding-param-data）入库
        '''
        binding='0|1|2|3.3|4.4|{}'.format(fp)
        # 成功入库，返回插入的记录index 
        rec_idx = vc.push_record(dbname, feat, binding)
        print("record insert OK! idx={}".format(rec_idx))
    
    print(dbname, vc.get_db_record_count(dbname))

    '''步骤3：检索'''
    with open('test.dat','rb') as f:
        raw_feat = f.read()
    # 使用ft进行特征规范化转换
    feat_to_retrieve = ft.tranform(raw_feat)
    # 特征检索获取记录。返回结果存到records中
    records = vc.retrieve_records(dbname, feat_to_retrieve)
    print("=======get retrieve results========")
    for r in records:
        print("\trec_idx={},similarity={},params={}".format(r[0],r[1],repr(r[2])))
    
    # 根据索引删除记录
    vc.delete_record(dbname, rec_idx)
    print("删除记录成功！")
    print(dbname, vc.get_db_record_count(dbname))
    # 数据库删除
    vc.delete_db(dbname)
    print("数据库删除成功！")


# coding=utf-8
"""
common API client component
for ISE /CnnExtractor /...
"""
import comm_util as _cu
import struct as _stu
import collections as _coll


class ApiError(Exception):
    """
        exception class for comm-Apis
    """

    def __init__(self, errc, errs):
        self.errc = errc
        self.errs = errs

    def __str__(self):
        return '{},{}'.format(self.errc, self.errs)

    def __repr__(self):
        return 'ApiError:' + self.__str__()


# Face Record for 3-elem part
FaceRecord3e = _coll.namedtuple('FaceRecord3e', ('idx', 'sim', 'sc_all', 'sc_top', 'sc_bot', 'params'))


class ApiCli:
    """
        'ISE'/'CnnExtractor' raw-tcp api
    """

    @staticmethod
    def __sock_recv_cmd(sock):
        d = _cu.sock_recv_all(sock, 4)
        if len(d) < 4:
            raise ApiError(-10, 'error recv cmd-len. recvbytes={}'.format(len(d)))
        cmdlen = _stu.unpack('<i', d)[0]

        if cmdlen < 0 or cmdlen > 100000000:
            raise ApiError(-1, "error parse 'cmdlen' ")
        rbt = _cu.sock_recv_all(sock, cmdlen)
        if len(rbt) < cmdlen:
            raise ApiError(-3, 'recv cmd partially! read=' + str(len(rbt)) + 'when req=' + str(cmdlen))
        return rbt

    ################################################################
    def __init__(self, addr=None, tmo=None):
        """
        Constructor
        """
        self.server_addr = addr or ('127.0.0.1', 2001)
        self.conn_tmo = tmo or 3000

    @staticmethod
    def __check_ret_err(sr: _cu.SeqReader2):
        retc = sr.rd_int()
        if retc != 0:
            errs = sr.rd_bstr()
            raise ApiError(retc, errs.decode())

    __tfm = {int: lambda val: (_stu.pack('<i', val),),
             float: lambda val: (_stu.pack('<f', val),),
             bytes: lambda val: (_stu.pack('<i', len(val)), val,)}

    def __api_cmd_comm(self, cmdid, *sargs):
        # type: (int, ...) -> _cu.SeqReader2
        sock = _cu.sock_conn_server(self.server_addr, self.conn_tmo)
        try:
            lst = []
            for a in sargs:
                lst.extend(self.__tfm[type(a)](a))
            content = b''.join(lst)
            bufcmd = b''.join((_stu.pack('<i', len(content) + 4), _stu.pack('<i', cmdid), content))
            sock.sendall(bufcmd)
            # sock.recv
            bufcmd = self.__sock_recv_cmd(sock)
            sr = _cu.SeqReader2(bufcmd)
            self.__check_ret_err(sr)
            return sr
        finally:
            sock.close()

    '''________________________________________'''

    def rt_get_all_db(self):
        sr = self.__api_cmd_comm(3)
        return [sr.rd_bstr().decode() for _ in range(sr.rd_int())]

    def rt_create_db(self, dbname):
        assert isinstance(dbname, str)
        self.__api_cmd_comm(1, dbname.encode())

    def rt_delete_db(self, dbname):
        assert isinstance(dbname, str)
        self.__api_cmd_comm(2, dbname.encode())

    def rt_push_face_fea(self, dbname, ftype, facefeat, paradata):
        assert isinstance(dbname, str) and isinstance(ftype, int) and \
               isinstance(facefeat, bytes) and isinstance(paradata, str)
        sr = self.__api_cmd_comm(10, dbname.encode(), ftype, facefeat, paradata.encode())
        return sr.rd_long()

    def rt_get_rec_ct(self, dbname):
        assert isinstance(dbname, str)
        sr = self.__api_cmd_comm(65, dbname.encode())
        return sr.rd_long()

    @staticmethod
    def __get_face_recs3(sr: _cu.SeqReader2):
        def _chkint(val, msg):
            if val < 0 or val > 1000000000:
                raise ApiError(-2, msg + str(val))

        retc = sr.rd_int()
        _chkint(retc, 'error get ImgRecord.count')

        def _readparam():
            pct = sr.rd_int()
            _chkint(pct, 'error get ImgRecord.param_ct ')
            return [sr.rd_bstr() for _i in range(pct)]

        return [FaceRecord3e(sr.rd_long(), sr.rd_float(), sr.rd_float(), sr.rd_float(), sr.rd_float(),
                             _readparam()) for _j in range(retc)]

    def rt_retrieve_face_3e(self, dbname, ftype, facefeat, where_stmt, min_sim=0.1, max_rec=8):
        # type: (str, int, bytes, str, float, int) -> list
        assert isinstance(dbname, str) and isinstance(ftype, int) and \
               isinstance(facefeat, bytes) and isinstance(where_stmt, str)
        sr = self.__api_cmd_comm(111, dbname.encode(), ftype, facefeat, where_stmt.encode(), max_rec, min_sim)
        return self.__get_face_recs3(sr)

    def rt_retrieve_face(self, dbname, facefeat, where_stmt, min_sim=0.1, max_rec=8):
        assert isinstance(dbname, str) and isinstance(facefeat, bytes) and isinstance(where_stmt, str)
        sr = self.__api_cmd_comm(11, dbname.encode(), 1, facefeat, where_stmt.encode(), max_rec, min_sim)
        def _chkint(val, msg):
            if val < 0 or val > 1000000000:
                raise ApiError(-2, msg + str(val))
        retc = sr.rd_int()
        _chkint(retc, 'error get ImgRecord.count')
        def _readparam():
            pct = sr.rd_int()
            _chkint(pct, 'error get ImgRecord.param_ct ')
            return [sr.rd_bstr() for _i in range(pct)]
        return [(sr.rd_long(), sr.rd_float(), _readparam()) for _j in range(retc)]
    
    def rt_test_io(self, dbname, chk_ct, io_ct):
        assert isinstance(dbname, str) and isinstance(chk_ct, int) and isinstance(io_ct, int)
        self.__api_cmd_comm(1200, dbname.encode(), chk_ct, io_ct)

    #############################################################

    def ff_extract_feat(self, ftype, imgdata):
        assert isinstance(ftype, int) and isinstance(imgdata, bytes)
        sr = self.__api_cmd_comm(802, ftype, imgdata)
        return sr.rd_bstr()

if __name__=='__main__':
    import async_util as _au
    import time,threading as _th
    pool = _au.TaskPool(8)
    lock = _th.Lock()
    cli = ApiCli(('127.0.0.1', 2021), 4000)
    sco = [0,0,0,0]
    t1=time.time()
    for _ in range(2000):
        def _task(sc):
            tc1=time.time()
            cli.rt_test_io('test_dilu_2kw',16,400)
            tu=time.time()-tc1
            with lock:
                sc[0]+=1
                sc[1]+=tu
        pool.push_task_fb(_task, sco)
    pool.join()
    t2=time.time()
    print("QPS=",sco[0]/(t2-t1), "avg-time(ms)=",1000.0*sco[1]/sco[0])   
    



# coding=utf-8
import ctypes as _ctp
import os as _os
import sys as _sys
import struct as _stu
from . import _comm_util as _cu
import collections as _coll
import threading as _th

__all__=['VseClient','VseError','FeatureTransformer','DataRecord']

assert _sys.version[0] == '3', 'python major version must be 3!'


#####################################################

def _s2b(strval):
    return strval.encode()

_dllname = 'vsepy_cimp'
_prefix = _os.path.split(_os.path.realpath(__file__))[0]
if _os.name in ("nt", "ce"):
    _dll = _ctp.CDLL(_prefix + '/{}.dll'.format(_dllname))

else:
    _dll = _ctp.CDLL(_prefix + '/lib{}.so'.format(_dllname))



_dll.transformer_init.restype = _ctp.c_void_p
_dll.transformer_transform.restype = _ctp.c_int32
_dll.transformer_get_param.restype = _ctp.c_int32

_SOCK_REUSE_FLG_DATA = _stu.pack('<i', 0x6666b003)

DataRecord = _coll.namedtuple('DataRecord', ('idx', 'sim', 'params'))

class VseError(Exception):
    """
        exception class for 'vse' sdk
        
        Attributes:
            errc: err-code (int)
            errs: err-message (string)
    """
    __slots__ = ('errc', 'errs')

    def __init__(self, errc, errs):
        self.errc = errc
        self.errs = errs

    def __str__(self):
        return repr(self.errc) + ',' + self.errs

    def __repr__(self):
        return 'VseError:' + self.__str__()


class FeatureTransformer(object):
    """特征转换类实例，提供vse所用特征的标准化转换"""
    __slots__ = ("__p",)
    __dl = _dll

    def __init__(self, client=None):
        # type: (str, VseClient) -> None
        """特征转换实例初始化。提供两种初始化方法：
            1.基于datapath初始化。用户需要指定正确的datapath，为包含与检索引擎相同的bin文件所在目录。
            2.基于client有效实例初始化。通过VseClient实例读取检索引擎所用的bin文件数据进行初始化（建议采用此方法）。
        """
        self.__p = None
        if client:
            assert isinstance(client, VseClient)
            data = client._get_trfm_param()
            p = self.__dl.transformer_init(data, _ctp.c_long(len(data)))
        else:
            raise Exception("no valid 'datapath' or 'client' param!")
        if not p:
            raise VseError(-1, 'c-impl: transformer_init failed')
        self.__p = _ctp.c_void_p(p)

    def __del__(self):
        if self.__p:
            self.__dl.transformer_free(self.__p)

    def __repr__(self):
        return 'FeatureTransformer({},{};{})@{}'.format(self.get_param(10),self.get_param(11),
            self.get_param(1),hex(self.__p.value))

    def get_param(self, pt):
        # type: (int) -> int
        return int( self.__dl.transformer_get_param(self.__p, _ctp.c_int32(pt)) )

    def tranform(self, rawfeat):
        # type: (bytes) -> bytes
        """执行特征转换（原生向量特征 -> 引擎标准化特征）
         rawfeat: 原生向量二进制数据(bytes格式)
        :return 标准化后的向量特征数据(bytes格式)，可用来入库、检索。
        """
        featlen = self.get_param(5)
        #print("featlen=",featlen)
        feat = b'\x00'*featlen
        rt = self.__dl.transformer_transform(self.__p, _ctp.c_char_p(rawfeat), _ctp.c_size_t(len(rawfeat)//4),
            feat)
        if rt:
            raise VseError(rt, b'input feature invalid')
        return feat
        

class _Long(int):
    pass

class VseClient(object):
    """Vse向量搜索引擎访问API类"""

    @staticmethod
    def __sock_recv_cmd(sock):
        d = _cu.sock_recv_all(sock, 4)
        if len(d) < 4:
            raise VseError(-10, 'error recv cmd-len. recvbytes={}'.format(len(d)))
        cmdlen = _stu.unpack('<i', d)[0]

        if cmdlen < 0 or cmdlen > 100000000:
            raise VseError(-1, "error parse 'cmdlen' ")
        rbt = _cu.sock_recv_all(sock, cmdlen)
        if len(rbt) < cmdlen:
            raise VseError(-3, 'recv cmd partially! read=' + str(len(rbt)) + 'when req=' + str(cmdlen))
        return rbt

    @staticmethod
    def __check_ret_err(sr: _cu.SeqReader2):
        retc = sr.rd_int()
        if retc != 0:
            errs = sr.rd_bstr()
            raise VseError(retc, errs.decode())

    __tfm = {int: lambda val: (_stu.pack('<i', val),),
             float: lambda val: (_stu.pack('<f', val),),
             _Long: lambda val: (_stu.pack('<q', val),),
             bytes: lambda val: (_stu.pack('<i', len(val)), val,)}

    def __api_cmd_comm(self, cmdid, *sargs):
        # type: (int, ...) -> _cu.SeqReader2
        #sock = _cu.sock_conn_server(self.server_addr, self.conn_tmo)
        sock = self.__pool_get_sock()
        try:
            lst = [_stu.pack('<i', cmdid)]
            for a in sargs:
                lst.extend(self.__tfm[type(a)](a))
            ctx_len = sum(map(len,lst))
            lst.insert(0, _stu.pack('<i', ctx_len))
            bufcmd = b''.join(lst)
            sock.sendall(bufcmd)
            # sock.recv
            bufcmd = self.__sock_recv_cmd(sock)
            try:
                sock.sendall(_SOCK_REUSE_FLG_DATA)
                self.__pool_put_sock(sock)
                sock = None
            except:
                print("_SOCK_REUSE_FLG_DATA send failed!")
            sr = _cu.SeqReader2(bufcmd)
            self.__check_ret_err(sr)
            return sr
        finally:
            if sock:
                sock.close()

        
    def __init__(self, addr, tmo=None):
        # type: (tuple, int|None) -> None
        """初始化实例并配置连接地址

         addr: 以(ip,port)形式的网络地址，如：('127.0.0.1', 2018)
         tmo: API访问连接超时时间（单位ms）。注意，VseClient实例下所有API均为阻塞模式工作，均支持多线程访问。
                当遇到网络拥塞或者业务执行过长情况，单个API可能执行时间长，可以通过此参数设置合理值。默认值为4000。
        """
        self.server_addr = addr or ('127.0.0.1', 2001)
        self.conn_tmo = tmo or 3000
        self._spool = _coll.deque(maxlen=16)
        self._lck = _th.Lock()

    def __pool_get_sock(self):
        with self._lck:
            if len(self._spool)>0:
                #print("reuse pop!!")
                return self._spool.pop()
        return _cu.sock_conn_server(self.server_addr, self.conn_tmo)
    
    def __pool_put_sock(self, sock):
        with self._lck:
            if len(self._spool)<self._spool.maxlen:
                self._spool.append(sock)
                #print("reuse push!!")
                return
        sock.close()

    def _get_trfm_param(self):
        # type: () -> bytes
        return self.__api_cmd_comm(1001).left_data()

    @staticmethod
    def __get_vse_recs(sr: _cu.SeqReader2):
        def _chkint(val, msg):
            if val < 0 or val > 1000000000:
                raise VseError(-2, msg + str(val))

        retc = sr.rd_int()
        _chkint(retc, 'error get ImgRecord.count')

        def _readparam():
            pct = sr.rd_int()
            _chkint(pct, 'error get ImgRecord.param_ct ')
            return [sr.rd_bstr() for _ in range(pct)]

        return [DataRecord(sr.rd_long(), sr.rd_float(), _readparam()) for _ in range(retc)]

    def enum_all_dbs(self):
        # type: () -> list[str]
        """枚举所有库
        :return 所有库名列表
        """
        sr = self.__api_cmd_comm(3)
        return [sr.rd_bstr().decode() for _ in range(sr.rd_int())]

    def create_db(self, dbname):
        # type: (str) -> None
        """新创建一个库
         dbname: 库名称（注意，库命名需要遵循变量命名规则，即：字母+数字+下划线，但是首字符需为字母）
        """
        assert isinstance(dbname,str)
        self.__api_cmd_comm(1, _s2b(dbname))
    
    def delete_db(self, dbname):
        # type: (str) -> None
        """删除一个库
         dbname: 库名称（注意，删除前需要保证没有其他用户或者线程对当前库进行操作，否则可能引发异常）
        """
        assert isinstance(dbname,str)
        self.__api_cmd_comm(2, _s2b(dbname))

    def get_db_record_count(self, dbname):
        # type: (str) -> int
        """获得指定库里记录个数
         dbname: 库名称
        :return 本库的记录个数
        """
        assert isinstance(dbname,str)
        sr = self.__api_cmd_comm(65, _s2b(dbname))
        return sr.rd_long()

    def push_record(self, dbname, feat, bindpara=''):
        # type: (str, bytes, str) -> int
        """向指定库插入记录（特征&绑定数据）
         dbname: 库名称
         feat: 待入库的标准化特征数据（需FeatureTransformer.transform后的结果，非原生特征数据！）
         bp: 绑定参数值列表（对于当前版本vse的绑定参数为定制化参数）
        :return 记录uid值(>=0，库表级全局唯一)
        注：当前vse版本支持的binding-data共6个字段，分别为：uint64,int64,int64,float32,float32,str(255)
            对于输入参数bp来说：1.可以留空，那么字段均按默认值(数字为0，字符串为空字串) 
            2.可以指定值，格式为以“|”符进行分隔的字符串。对于不指定的位置可以留空。如：
                '0|1|2|3.3|4.4|testhhh' 或 '|87347||5e3||'
        """
        assert isinstance(dbname, str) and isinstance(feat, bytes)
        sr = self.__api_cmd_comm(10, _s2b(dbname), 1, feat, _s2b(bindpara))
        return sr.rd_long()
    
    def retrieve_records(self, dbname, feat, wherestmt='', min_sim=0.1, max_rec=8):
        # type: (str, bytes, str, float, int) -> list[tuple]
        """在指定库执行向量相似检索，返回按相似度（从高到低）排序后的结果
         dbname: 库名称
         feat: 待检索的标准化特征数据（需FeatureTransformer.transform后的结果，非原生特征数据！）
         min_sim: 最小相似度阈值。特征相似度小于此阈值的结果将被过滤。
         max_rec: 最大返回结果数。基于相似度排序和阈值过滤后结果数如果超过此值，将截断。
        :return 检索结果记录的列表。外部容器为list，每条结果为一个tuple，具体格式为：(记录uid,相似度值,参数列表)。每个子项内容为:
            :记录uid：同push_record返回值,delete_record参数值。为系统内分配的记录唯一id
            :相似度值：值域为[0,1.0]的相似度得分值。对于完全相同的向量相似度为1，完全不相同的向量相似度为0
            :参数列表：格式为list[bytes]的binding-data数据。与push_record插入到记录中的数据对应。对于当前版本vse，binding-data数量为6，格式可见push_record说明。
        
        """
        assert isinstance(dbname, str) and isinstance(feat, bytes)
        sr = self.__api_cmd_comm(11, _s2b(dbname), 1, feat, _s2b(wherestmt), max_rec, float(min_sim))
        return self.__get_vse_recs(sr)

    def delete_record(self, dbname, ridx):
        # type: (str, int) -> None
        """
          根据索引删除记录
          @param: dbname 数据库名
          @param : ridx 待删除记录索引值
        """
        assert isinstance(dbname, str) and isinstance(ridx, int)
        self.__api_cmd_comm(14, _s2b(dbname), _Long(ridx))
    
    def delete_record_ws(self, dbname, ws):
        # type: (str, str) -> None
        assert isinstance(dbname, str) and isinstance(ws, str) 
        self.__api_cmd_comm(21, _s2b(dbname), _s2b(ws))

    def rt_test_io(self, dbname, chk_ct, io_ct):
        assert isinstance(dbname, str) and isinstance(chk_ct, int) and isinstance(io_ct, int)
        self.__api_cmd_comm(1200, dbname.encode(), chk_ct, io_ct)

    def scan_repair1(self, dbname):
        assert isinstance(dbname, str) 
        return self.__api_cmd_comm(1101, _s2b(dbname)).rd_int()


'''testing code'''

if __name__=='__main__':
    l= _Long(111)
    print(type(l))


# coding=utf-8
import struct as _stu
import sys as _sys
import socket as _sock

if _sys.version[0] == '2':
    from cStringIO import _BytesIO

    def _s2b(sv):
        return sv.encode() if isinstance(sv, unicode) else sv
else:
    from io import BytesIO as _BytesIO

    def _s2b(sv):
        return sv if isinstance(sv, bytes) else sv.encode()


class SeqReader2(object):
    def __init__(self, dat: bytes, off=0):
        assert isinstance(dat, bytes)
        self.__bts = dat
        self.__off = off

    offset = property(lambda self: self.__off,
                      doc='data sequence read offset')

    left_bytes = property(lambda self: len(self.__bts) - self.__off,
                          doc='data sequence left bytes that can be read')

    def __readtype(self, typelen, fmt):
        tmp = self.__bts[self.__off: self.__off + typelen]
        self.__off += typelen
        return _stu.unpack(fmt, tmp)

    def read_types(self, fmts):
        sz = _stu.calcsize(fmts)
        tmp = self.__bts[self.__off: self.__off + sz]
        self.__off += sz
        return _stu.unpack(fmts, tmp)

    def read_bytes(self, sz):
        tmp = self.__bts[self.__off: self.__off + sz]
        self.__off += sz
        return tmp

    def rd_bstr(self):
        sz = self.rd_int()
        return self.read_bytes(sz)

    def rd_int(self):
        return self.__readtype(4, '<I')[0]

    def rd_long(self):
        return self.__readtype(8, '<q')[0]

    def rd_float(self):
        return self.__readtype(4, '<f')[0]


def sock_recv_all(sock: _sock.socket, rcvlen: int):
    tmpbuf, off = _BytesIO(), 0
    try:
        while off < rcvlen:
            rbt = sock.recv(min(rcvlen - off, 1024 * 1024))
            if len(rbt) == 0:
                break
            tmpbuf.write(rbt)
            off += len(rbt)
        return tmpbuf.getvalue()
    finally:
        tmpbuf.close()


def sock_conn_server(server_addr: tuple, conn_tmo_ms: float) -> _sock.socket:
    s = None
    try:
        s = _sock.create_connection(server_addr, conn_tmo_ms / 1000.0)
        s.setsockopt(_sock.IPPROTO_TCP, _sock.TCP_NODELAY, 1)
        s.setblocking(True)
        return s
    except:
        if s:
            s.close()
        raise


"""
class PutChsText(object):
    def __init__(self, ttf):
        self._face = _ft.Face(ttf)
        #hscale = 1.0
        #matrix = _ft.Matrix(int(hscale)*0x10000, int(0.2*0x10000), \
        #                    int(0.0*0x10000), int(1.1*0x10000))
        #pen_translate = _ft.Vector()
        #self._face.set_transform(matrix, pen_translate)

    def draw_text(self, image, pos, text, text_size, text_color):
        '''
        draw chinese(or not) text with ttf
        :param image:     image(numpy.ndarray) to draw text
        :param pos:       where to draw text
        :param text:      the context, for chinese should be unicode type
        :param text_size: text size
        :param text_color:text color
        :return:          image
        '''
        self._face.set_char_size(text_size * 64)
        metrics = self._face.size
        ascender = metrics.ascender/64.0

        #descender = metrics.descender/64.0
        #height = metrics.height/64.0
        #linegap = height - ascender + descender
        ypos = int(ascender)

        if not isinstance(text, str):
            text = text.decode('utf-8')
        img = self.draw_string(image, pos[0], pos[1]+ypos, text, text_color)
        return img

    def draw_string(self, img, x_pos, y_pos, text, color):
        '''
        draw string
        :param x_pos: text x-postion on img
        :param y_pos: text y-postion on img
        :param text:  text (unicode)
        :param color: text color
        :return:      image
        '''
        prev_char = 0
        px = x_pos << 6   # div 64
        py = y_pos << 6

        cur_pen = _ft.Vector()

        for cur_char in text:
            self._face.load_char(cur_char)
            kerning = self._face.get_kerning(prev_char, cur_char)
            px += kerning.x
            slot = self._face.glyph

            cur_pen.x = px
            cur_pen.y = py - slot.bitmap_top * 64
            self.draw_ft_bitmap(img, slot.bitmap, cur_pen, color)

            px += slot.advance.x
            prev_char = cur_char

        return img

    def draw_ft_bitmap(self, img, bitmap, pen, color):
        # type: (_np.ndarray, _ft.Bitmap, _ft.Vector(), ()) -> None
        '''
        draw each char
        :param bitmap: bitmap
        :param pen:    pen
        :param color:  pen color e.g.(0,0,255) - red
        :return:       image
        '''
        x_pos = pen.x >> 6
        y_pos = pen.y >> 6
        cols = bitmap.width
        rows = bitmap.rows

        glyph_pixels = bitmap.buffer

        for row in range(rows):
            for col in range(cols):
                cp = glyph_pixels[row*cols + col]
                if cp != 0 and y_pos+row < img.shape[0] \
                        and x_pos+col < img.shape[1]:
                    pix = img[y_pos + row][x_pos + col]
                    pix[0] = (color[0]*cp+pix[0]*(255-cp))/255
                    pix[1] = (color[1]*cp+pix[1]*(255-cp))/255
                    pix[2] = (color[2]*cp+pix[2]*(255-cp))/255
"""



from distutils.core import setup
from setuptools import find_packages

with open("README.rst", "r") as f:
    long_description = f.read()

setup(
    name='pytensordb',
    version='1.0.0',
    description='TensorDB API(Python)',
    long_description=long_description,
    author='上海爱可生信息技术股份有限公司',
    author_email='zhoupeng@actionsky.com',
    install_requires=[],
    url='https://www.actionsky.com/Cloud-tree-series/tensordb.html',
    packages=find_packages(),
    include_package_data=True,
    license='Apache LICENSE 2.0',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries'
    ],
)
